#include <iostream>
#include <blt/std/allocator.h>
#include <blt/gfx/window.h>
#include <blt/gfx/state.h>
#include <blt/gfx/stb/stb_image.h>
#include <functions.h>
#include "blt/gfx/renderer/resource_manager.h"
#include "blt/gfx/renderer/batch_2d_renderer.h"
#include "blt/std/assert.h"
#include "imgui.h"
#include "extern.h"
#include <variant>
#include <random>
#include <queue>
#include <stack>
#include <config.h>
#include <gp.h>

blt::gfx::matrix_state_manager global_matrices;
blt::gfx::resource_manager resources;
blt::gfx::batch_renderer_2d renderer_2d(resources);

class tree
{
    public:
        struct crossover_result_t
        {
            std::unique_ptr<tree> c1, c2;
        };
    private:
        std::unique_ptr<node, node_deleter> root = nullptr;
        
        explicit tree(node* root): root(root)
        {}
        
        void normalize_image()
        {
            float mx = 0, my = 0, mz = 0;
            float sx = 0, sy = 0, sz = 0;
            for (auto& v : root->getImage().getData())
            {
                for (int i = 0; i < 3; i++)
                {
                    if (std::isnan(v[i]))
                    {
                        v[i] = 0.0f;
                    }
                }
                mx = std::max(v.x(), mx);
                my = std::max(v.y(), my);
                mz = std::max(v.z(), mz);
                sx = std::min(v.x(), sx);
                sy = std::min(v.y(), sy);
                sz = std::min(v.z(), sz);
            }
            for (auto& v : root->getImage().getData())
            {
                if (mx - sx != 0)
                    v[0] = (v.x() - sx) / (mx - sx);
                if (my - sy != 0)
                    v[1] = (v.y() - sy) / (my - sy);
                if (mz - sz != 0)
                    v[2] = (v.z() - sz) / (mz - sz);
            }
        }
        
        node* select_random_child()
        {
            static std::random_device dev;
            static std::mt19937_64 engine{dev()};
            auto d = depth();
            std::uniform_int_distribution depth_dist(0, static_cast<int>(d));
            std::uniform_int_distribution select(0, 3);
            
            node* current = root.get();
            
            while (true)
            {
                std::uniform_int_distribution<size_t> children(0, current->argc - 1);
            }
        }
    
    public:
        static std::unique_ptr<tree> construct_random_tree()
        {
            return std::make_unique<tree>(tree{node::construct_random_tree()});
        }
        
        blt::size_t depth()
        {
            // depth -> node
            std::stack<std::pair<blt::size_t, node*>> stack;
            blt::size_t max_depth = 0;
            
            stack.emplace(0, root.get());
            while (!stack.empty())
            {
                auto top = stack.top();
                auto* node = top.second;
                auto depth = top.first;
                max_depth = std::max(max_depth, depth);
                for (blt::size_t i = 0; i < node->argc; i++)
                {
                    if (node->sub_nodes[i] != nullptr)
                    {
                        stack.emplace(depth + 1, node->sub_nodes[i]);
                    }
                }
                stack.pop();
            }
            
            return max_depth;
        }
        
        static crossover_result_t crossover(tree* p1, tree* p2)
        {
            return {};
        }
        
        void evaluate()
        {
            root->evaluate_tree();
            normalize_image();
        }
        
        image& getImage()
        {
            return root->getImage();
        }
        
        bool hasImage()
        {
            return root->hasImage();
        }
        
        bool hasTree()
        {
            return root != nullptr;
        }
        
        float fitness()
        {
            return eval_DNF_SW(root->getImage());
        }
        
        void printTree()
        {
            root->printTree();
        }
};

std::unique_ptr<tree> test_tree;
blt::gfx::texture_gl2D* texture;

void print_bits(float f)
{
    auto u = blt::mem::type_cast<blt::u32>(f);
    for (size_t i = 31; i > 0; i--)
    {
        std::cout << ((u >> i) & 0x1);
        if (i % 8 == 0)
            std::cout << ", ";
        else if (i % 4 == 0)
            std::cout << " ";
    }
    std::cout << std::endl;
}

void init()
{
    global_matrices.create_internals();
    renderer_2d.create();
    
    texture = new blt::gfx::texture_gl2D(width, height);
    resources.set("img", texture);
    resources.enqueue("../libraries/BLT-With-Graphics-Template/resources/textures/parkerfemBOY.png", "cum");
    
    resources.load_resources();
}

void update(std::int32_t w, std::int32_t h)
{
    global_matrices.update_perspectives(w, h, 90, 0.1, 2000);
    glDisable(GL_BLEND);
    
    if (ImGui::Button("Regenerate"))
    {
        BLT_INFO("Regen tree");
        test_tree = tree::construct_random_tree();
        test_tree->evaluate();
        BLT_INFO("Uploading");
        texture->upload((void*) test_tree->getImage().getData().data(), GL_RGB, 0, 0, 0, -1, -1, GL_FLOAT);
    }
    
    if (ImGui::Button("Display"))
    {
        if (test_tree->hasTree())
            test_tree->printTree();
    }
    
    if (ImGui::Button("Eval"))
    {
        if (test_tree->hasImage())
            BLT_DEBUG(test_tree->fitness());
    }
    
    auto lw = 512.0f;
    auto lh = 512.0f;
    //renderer_2d.drawRectangle(blt::vec4{0.5, 0.0, 1.0, 1.0},
    //                          {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, static_cast<float>(w), static_cast<float>(h)});
    renderer_2d.drawRectangle("img", {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, lw, lh});
    
    global_matrices.update();
    renderer_2d.render();
}

int main()
{
    shapiro_test_run();
    blt::gfx::init(blt::gfx::window_data{"Window of GP test", init, update}.setSyncInterval(1));
    global_matrices.cleanup();
    resources.cleanup();
    renderer_2d.cleanup();
    blt::gfx::cleanup();
    return 0;
}
